/*
** nester - NES emulator
** Copyright (C) 2000  Darren Ranalli
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful, 
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
** Library General Public License for more details.  To obtain a 
** copy of the GNU Library General Public License, write to the Free 
** Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
** Any permitted reproduction of these routines, in whole or in part,
** must bear this legend.
*/
#include <windows.h>
#include <shlwapi.h>
#include <string.h>
#include <math.h>
#include "NES.h"
#include "NES_screen_mgr.h"
#include "NES_ROM.h"
#include "NES_PPU.h"
#include "pixmap.h"
#include "SNSS.h"
#include "mkutils.h"

#include "debug.h"

const uint8 NES::NES_preset_palette[NES_NUM_COLORS][3] = 
{
// include the NES palette
#include "NES_pal.h"
};

//const float NES::CYCLES_PER_LINE = (float)113.6;
//const float NES::CYCLES_PER_LINE = (float)113.852;
//const float NES::CYCLES_PER_LINE = (float)113.75;
//const float NES::CYCLES_PER_LINE = (float)113.6666666666666666666;

// External Device
#define EX_NONE                   0
#define EX_ARKANOID               2
#define EX_CRAZY_CLIMBER          3
#define EX_DATACH_BARCODE_BATTLER 4
#define EX_DOREMIKKO_KEYBOARD     5
#define EX_EXCITING_BOXING        6
#define EX_FAMILY_KEYBOARD        7
#define EX_FAMILY_TRAINER_A       8
#define EX_FAMILY_TRAINER_B       9
#define EX_HYPER_SHOT             10
#define EX_MAHJONG                11
#define EX_OEKAKIDS_TABLET        12
#define EX_OPTICAL_GUN            13
#define EX_POKKUN_MOGURAA         14
#define EX_POWER_PAD_A            15
#define EX_POWER_PAD_B            16
#define EX_TOP_RIDER              17
#define EX_TURBO_FILE             18

NES::NES(const char* ROM_name, NES_screen_mgr* _screen_mgr, sound_mgr* _sound_mgr, HWND parent_window_handle)
{
  scr_mgr = _screen_mgr;
  snd_mgr = _sound_mgr;

  main_window_handle = parent_window_handle;

  scr_mgr->setParentNES(this);

  LOG("nester - NES emulator by Darren Ranalli, (c) 2000\n");

  cpu = NULL;
  ppu = NULL;
  apu = NULL;

  try {
    LOG("Creating NES CPU...");
    cpu = new NES_6502(this);
    if(!cpu) throw "error allocating cpu";
    LOG("Done.\n");

    LOG("Creating NES PPU...");
    ppu = new NES_PPU(this);
    if(!ppu) throw "error allocating ppu";
    LOG("Done.\n");

    LOG("Creating NES APU...");
    apu = new NES_APU(this);
    if(!apu) throw "error allocating apu";
    LOG("Done.\n");

    loadROM(ROM_name);
  } catch(...) {
    if(cpu) delete cpu;
    if(ppu) delete ppu;
    if(apu) delete apu;
    throw;
  }

  // set up palette and assert it
  use_vs_palette = 0;
  calculate_palette();
  scr_mgr->assert_palette();

  pad1 = NULL;
  pad2 = NULL;

  is_frozen = FALSE;
}

NES::~NES()
{
  freeROM();

  if(cpu) delete cpu;
  if(ppu) delete ppu;
  if(apu) delete apu;
}

void NES::new_snd_mgr(sound_mgr* _sound_mgr)
{
  snd_mgr = _sound_mgr;
  apu->snd_mgr_changed();
}

void NES::loadROM(const char* fn)
{
  LOG("Loading ROM...");

  ROM = new NES_ROM(fn);

  // set up the mapper
  mapper = GetMapper(this, ROM);
  if(!mapper)
  {
    // unsupported mapper
    LOG("mapper #" << (int)ROM->get_mapper_num() << " not supported" << endl);

    delete ROM;
    ROM = NULL;

    throw "unsupported mapper";
  }

  LOG("Done\n");

  LOG(ROM->GetRomName() << ".nes: #" << (int)ROM->get_mapper_num() << " ");
  switch(ROM->get_mirroring())
  {
    case NES_PPU::MIRROR_HORIZ:
      LOG("H ");
      break;
    case NES_PPU::MIRROR_VERT:
      LOG("V ");
      break;
    case NES_PPU::MIRROR_FOUR_SCREEN:
      LOG("F ");
      break;
  }

  if(ROM->has_save_RAM())
  {
    LOG("S ");
  }
  if(ROM->has_trainer())
  {
    LOG("T ");
  }

  LOG(16*ROM->get_num_16k_ROM_banks() << "K/" << 8*ROM->get_num_8k_VROM_banks() << "K " << endl);

  // load datas to save it at the top of reset()
  Load_SaveRAM();
  Load_Disk();

  reset();

  LOG("Starting emulation...\n");
}

void NES::freeROM()
{
  Save_SaveRAM();
  Save_Disk();
  Save_TurboFile();

  LOG("Freeing ROM...");
  if(ROM)
  {
    delete ROM;
    ROM = NULL;
  }
  if(mapper)
  {
    delete mapper;
    mapper = NULL;
  }

  scr_mgr->clear(0x00);
  scr_mgr->blt();

  LOG("Done\n");
  LOG(endl);
}

const char* NES::getROMname()
{
  return ROM->GetRomName();
}

const char* NES::getROMpath()
{
  return ROM->GetRomPath();
}

void NES::reset()
{
  // save SaveRAM
  Save_SaveRAM();
  Save_Disk();
  Save_TurboFile();

  // clear RAM
  memset(RAM, 0x00, sizeof(RAM));

  // set up CPU
  {
    NES_6502::Context context;

    memset((void*)&context, 0x00, sizeof(context));
    cpu->GetContext(&context);

    context.mem_page[0] = RAM;
    context.mem_page[3] = SaveRAM;

    cpu->SetContext(&context);
  }

  // reset the PPU
  ppu->reset();

  // VROM write protect
  if(ROM->get_num_8k_VROM_banks())
  {
    ppu->vram_write_protect = 1;
  }
  else
  {
    ppu->vram_write_protect = 0;
  }

  // reset the APU
  apu->reset();

  frame_irq_enabled = 0xFF;
  frame_irq_disenabled = 0;

  if(mapper)
  {
    // reset the mapper
    mapper->Reset();
  }

  // reset the CPU
  cpu->Reset();

  // load SaveRAM
  Load_SaveRAM();
  Load_Disk();
  Load_TurboFile();

  // set up the trainer if present
  if(ROM->has_trainer())
  {
    // trainer is located at 0x7000; SaveRAM is 0x2000 bytes at 0x6000
    memcpy(&SaveRAM[0x1000], ROM->get_trainer(), NES_ROM::TRAINER_LEN);
  }

  uint32 j, sum_16k_rom = 0;
  uint8* ROM_banks = ROM->get_ROM_banks();
  for(j = 0; j < 8192; j++) sum_16k_rom += ROM_banks[j];
  for(j = 0; j < 8192; j++) sum_16k_rom += ROM_banks[8192*(2*ROM->get_num_16k_ROM_banks()-1)+j];
  sum_16k_rom &= 0xFFFFFF;

  #include "NES_set_Controller.cpp"
  #include "NES_set_VSPalette.cpp"
  #include "NES_set_Cycles.cpp"

  ideal_cycle_count  = 0.0;
  emulated_cycle_count = 0;

  pad_strobe = FALSE;
  pad1_bits = 0x00;
  pad2_bits = 0x00;

  kb_graph = 0;

  tf_pre_flag = 0x00;
  tf_pointer = 0x00;
  tf_bit = 0x01;

  if(ex_controller_type == EX_OEKAKIDS_TABLET)
  {
    POINT p;
    p.x = 16;
    p.y = 16;
    ClientToScreen (main_window_handle, &p);
    SetCursorPos ( p.x, p.y );
  }

  reg4016 = reg4016s3 = reg4016s4 = 0x00;
  reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0x00;
}

boolean NES::emulate_frame(boolean draw)
{
  uint32 i;
  pixmap p;
  uint8* cur_line; // ptr to screen buffer
  boolean retval = draw;

  trim_cycle_counts();

  // do frame
  ppu->start_frame();

  if(retval)
  {
    if(!scr_mgr->lock(p))
      retval = FALSE;
    else
      cur_line = p.data;
  }

  // LINES 0-239
  for(i = 0; i < NES_NUM_FRAME_LINES; i++)
  {
    // do one line's worth of CPU cycles
    //emulate_CPU_cycles(CYCLES_PER_LINE);
    //mapper->HSync(i);

    if(retval)
    {
      if(BANKSWITCH_PER_TILE)
      {
        // render line
        ppu->do_scanline_and_draw(cur_line, CYCLES_PER_LINE * 32 / 42);
        // do half line's worth of CPU cycles (hblank)
        emulate_CPU_cycles(13);
        mapper->HSync(i);
        emulate_CPU_cycles(CYCLES_PER_LINE * 10 / 42 - 13);
        if(i == 0)
        {
          emulate_CPU_cycles(CYCLES_PER_LINE * 32 / 32 + 13);
          mapper->HSync(i);
          emulate_CPU_cycles(CYCLES_PER_LINE * 10 / 42 - 13);
        }
      }
      else
      {
        // do one line's worth of CPU cycles
        emulate_CPU_cycles(CYCLES_PER_LINE);
        mapper->HSync(i);
        // render line
        ppu->do_scanline_and_draw(cur_line, 0);
      }
      // point to next line
      cur_line += p.pitch;
    }
    else
    {
      // do one line's worth of CPU cycles
      emulate_CPU_cycles(CYCLES_PER_LINE);
      mapper->HSync(i);
      ppu->do_scanline_and_dont_draw();
    }
  }

  if(retval)
  {
    scr_mgr->unlock();
  }

  ppu->end_frame();

  // fram_IRQ
  if(!(frame_irq_enabled & 0xC0))
  {
    cpu->DoIRQ();
  }

  for(i = 240; i <= 262; i++) // 261->262
  {
    if(i == 243) // 241->243
    {
      // do v-blank
      ppu->start_vblank();
      mapper->VSync();
    }
    else if(i == 262) // 261->262
    {
      ppu->end_vblank();
    }

    if(i == 243) // 241->243
    {
      // 1 instruction between vblank flag and NMI
      emulate_CPU_cycles(CYCLES_BEFORE_NMI);
      if(ppu->NMI_enabled()) cpu->DoNMI();
      emulate_CPU_cycles(CYCLES_PER_LINE - CYCLES_BEFORE_NMI);
      mapper->HSync(i);
      continue;
    }

    emulate_CPU_cycles(CYCLES_PER_LINE);
    mapper->HSync(i);
  }

  // HALF-LINE 262.5
  emulate_CPU_cycles(CYCLES_PER_LINE/2);

  apu->DoFrame();
  apu->SyncAPURegister();

  return retval;
}

void NES::freeze()
{
  apu->freeze();
  is_frozen = TRUE;
}

void NES::thaw()
{
  apu->thaw();
  is_frozen = FALSE;
}

boolean NES::frozen()
{
  return is_frozen;
}

uint8 NES::MemoryRead(uint32 addr)
{
//  LOG("Read " << HEX(addr,4) << endl);

  if(addr < 0x2000) // RAM
  {
    return ReadRAM(addr);
  }
  else if(addr < 0x4000) // low registers
  {
    return ReadLowRegs(addr);
  }
  else if(addr < 0x4018) // high registers
  {
    return ReadHighRegs(addr);
  }
  else if(addr < 0x6000) // mapper low
  {
//    LOG("MAPPER LOW READ: " << HEX(addr,4) << endl);
//    return((uint8)(addr >> 8)); // what's good for conte is good for me
    if(ROM->is_VSUnisystem())
    {
      if(addr == 0x5E00)
      {
        vs_index = 0;
        return 0xFF;
      }
      else if(addr == 0x5E01)
      {
        switch(vs_index++)
        {
          case 0x09:
            return 0x6F;
          default:
            return 0xB4;
        }
      }
      else
      {
        return mapper->MemoryReadLow(addr);
      }
    }
    return mapper->MemoryReadLow(addr);
  }
  else // save RAM, or ROM (mapper 40)
  {
    mapper->MemoryRead(addr);
    return cpu->GetByte(addr);
  }
}

void NES::MemoryWrite(uint32 addr, uint8 data)
{
//  LOG("Write " << HEX(addr,4) << " " << HEX(data,2) << endl);

  if(addr < 0x2000) // RAM
  {
    WriteRAM(addr, data);
  }
  else if(addr < 0x4000) // low registers
  {
    WriteLowRegs(addr, data);
    //mapper->WriteLowRegs(addr, data);
  }
  else if(addr < 0x4018) // high registers
  {
    WriteHighRegs(addr, data);
    mapper->WriteHighRegs(addr, data);
  }
  else if(addr < 0x6000) // mapper low
  {
    mapper->MemoryWriteLow(addr, data);
  }
  else if(addr < 0x8000) // save RAM
  {
    SaveRAM[addr - 0x6000] = data;
    mapper->MemoryWriteSaveRAM(addr, data);
  }
  else // mapper
  {
    mapper->MemoryWrite(addr, data);
  }
}


uint8 NES::ReadRAM(uint32 addr)
{
  return RAM[addr & 0x7FF];
}

void NES::WriteRAM(uint32 addr, uint8 data)
{
  RAM[addr & 0x7FF] = data;
}


uint8 NES::ReadLowRegs(uint32 addr)
{
  return ppu->ReadLowRegs(addr & 0xE007);
}

void NES::WriteLowRegs(uint32 addr, uint8 data)
{
  ppu->WriteLowRegs(addr & 0xE007, data);
}


uint8 NES::ReadHighRegs(uint32 addr)
{
  if(addr == 0x4014) // SPR-RAM DMA
  {
    LOG("Read from SPR-RAM DMA reg??" << endl);
    return ppu->Read0x4014();
  }
  else if(addr == 0x4015 && !(frame_irq_enabled & 0xC0)) // frame_IRQ
  {
    return apu->Read(0x4015) | 0x40;
  }
  else if(addr < 0x4016) // APU
  {
//    LOG("APU READ:" << HEX(addr,4) << endl);
    return apu->Read(addr);
  }
  else // joypad regs
  {
    uint8 retval;

    if(addr == 0x4016)
    {
      // joypad 1
      retval =  pad1_bits & 0x01;
      retval |= reg4016;
      //retval |= (reg4016s3 & 0x01) << 3;
      //retval |= (reg4016s4 & 0x01) << 4;
      pad1_bits >>= 1;
      //reg4016s3 >>= 1;
      //reg4016s4 >>= 1;
    }
    else
    {
      // Family Basic Key Board (thanx rinao)
      if(ex_controller_type == EX_FAMILY_KEYBOARD)
      {
        if(GetAsyncKeyState(VK_NEXT) & 0x8000) kb_graph = 1;
        if(GetAsyncKeyState(VK_PRIOR) & 0x8000) kb_graph = 0;

        reg4017 = 0xFF;
        if(kb_out)
        {
          switch(kb_scan)
          {
            case 1:
              {
                if(GetAsyncKeyState(VK_F8) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState(VK_RETURN) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('[' + 0x80) & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState(']' + 0x80) & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 2:
              {
                if(GetAsyncKeyState(VK_F7) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('@' + 0x80) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState(':' + 0x80) & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState(';' + 0x80) & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 3:
              {
                if(GetAsyncKeyState(VK_F6) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('O') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('L') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('K') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 4:
              {
                if(GetAsyncKeyState(VK_F5) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('I') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('U') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('J') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 5:
              {
                if(GetAsyncKeyState(VK_F4) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('Y') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('G') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('H') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 6:
              {
                if(GetAsyncKeyState(VK_F3) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('T') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('R') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('D') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 7:
              {
                if(GetAsyncKeyState(VK_F2) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('W') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('S') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('A') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 8:
              {
                if(GetAsyncKeyState(VK_F1) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState(VK_ESCAPE) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('Q') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState(VK_CONTROL)) reg4017 &= 0xEF;
              }
              break;

            case 9:
              {
                if(GetAsyncKeyState(VK_HOME) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState(VK_UP) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState(VK_RIGHT) & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState(VK_LEFT) & 0x8000) reg4017 &= 0xEF;
              }
              break;
          }
        }
        else
        {
          switch(kb_scan)
          {
            case 1:
              {
                if(GetAsyncKeyState(VK_F12) & 0x8000) reg4017 &= 0xFD;
                //if(GetAsyncKeyState(VK_SHIFT)) reg4017 &= 0xFB;  //shift(right)
                if(GetAsyncKeyState('\\' + 0x80) & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState(VK_END) & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 2:
              {
                if(GetAsyncKeyState(226) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('/' + 0x90) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('-' + 0x90) & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('^' + 0x80) & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 3:
              {
                if(GetAsyncKeyState('.' + 0x90) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState(',' + 0x90) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('P') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('0') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 4:
              {
                if(GetAsyncKeyState('M') & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('N') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('9') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('8') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 5:
              {
                if(GetAsyncKeyState('B') & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('V') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('7') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('6') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 6:
              {
                if(GetAsyncKeyState('F') & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('C') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('5') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('4') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 7:
              {
                if(GetAsyncKeyState('X') & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState('Z') & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState('E') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('3') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 8:
              {
                if(GetAsyncKeyState(VK_SHIFT)) reg4017 &= 0xFD;
                //if(GetAsyncKeyState(VK_MENU)) reg4017 &= 0xFB;
                if(kb_graph) reg4017 &= 0xFB;
                if(GetAsyncKeyState('1') & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState('2') & 0x8000) reg4017 &= 0xEF;
              }
              break;

            case 9:
              {
                if(GetAsyncKeyState(VK_DOWN) & 0x8000) reg4017 &= 0xFD;
                if(GetAsyncKeyState(VK_SPACE) & 0x8000) reg4017 &= 0xFB;
                if(GetAsyncKeyState(VK_DELETE) & 0x8000) reg4017 &= 0xF7;
                if(GetAsyncKeyState(VK_INSERT) & 0x8000) reg4017 &= 0xEF;
              }
              break;
          }
        }
      }
      // Optical Gun (Zapper)
      if(ex_controller_type == EX_OPTICAL_GUN)
      {
        reg4017 = 0x08;
        if(GetAsyncKeyState(VK_LBUTTON) & 0x8000) reg4017 |= 0x10;
        if(GetAsyncKeyState(VK_RBUTTON) & 0x8000) reg4017 |= 0x10;
        try
        {
          POINT pt;
          GetCursorPos(&pt);
          ScreenToClient(main_window_handle, &pt);
          if(NESTER_settings.nes.graphics.osd.double_size)
          {
            pt.x >>= 1;
            pt.y >>= 1;
          }
          if(!NESTER_settings.nes.graphics.show_all_scanlines)
          {
            pt.y += 8;
          }
          if(pt.x == 0) pt.x = 1;
          if(pt.x == 255) pt.x = 254;
          if(pt.y == 0) pt.y = 1;
          if(pt.y == 239) pt.x = 238;

          if(0 < pt.x && pt.x < 255 && 0 < pt.y && pt.y < 239)
          {
            for(int i = -1; i <= 1; i++)
            {
              for(int j = -1; j <= 1; j++)
              {
                uint8 c = ppu->GetPointColor(pt.x+i, pt.y+j) & 0x3F;
                if(NES_RGB_pal[c][0] == 0xFF && NES_RGB_pal[c][1] == 0xFF && NES_RGB_pal[c][2] == 0xFF)
                {
                  reg4017 &= 0xF7;
                }
                // for Chiller(#11)
                /*
                if(NES_RGB_pal[c][0] > 0 && NES_RGB_pal[c][1] == 0 && NES_RGB_pal[c][2] > 0)
                {
                  reg4017 &= 0xF7;
                }
                */
              }
            }
          }
        } catch(...) { }
      }
      // joypad 2
      retval =  pad2_bits & 0x01;
      retval |= reg4017;
      retval |= (reg4017s1 & 0x01) << 1;
      retval |= (reg4017s3 & 0x01) << 3;
      retval |= (reg4017s4 & 0x01) << 4;
      pad2_bits >>= 1;
      reg4017s1 >>= 1;
      reg4017s3 >>= 1;
      reg4017s4 >>= 1;
    }
    return retval;
  }
}

void NES::WriteHighRegs(uint32 addr, uint8 data)
{
  if(addr == 0x4014) // SPR-RAM DMA
  {
    ppu->Write0x4014(data);
    cpu->SetDMA(CYCLES_PER_DMA);
  }
  else if(addr < 0x4016) // APU
  {
    apu->Write(addr, data);
  }
  else if(addr == 0x4016) // joy pad
  {
    // bit 0 == joypad strobe
    if(data & 0x01)
    {
      pad_strobe = TRUE;
    }
    else
    {
      if(pad_strobe)
      {
        pad_strobe = FALSE;
        reg4016 = reg4016s3 = reg4016s4 = 0;
        reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;

        // get input states
        if(pad1) pad1_bits = pad1->get_inp_state();
        if(pad2) pad2_bits = pad2->get_inp_state();
        // for VS Super Sky Kid, VS Dr.Mario
        if(pad_swap)
        {
          pad_swap = pad1_bits;
          pad1_bits = (pad1_bits & 0x0C) | (pad2_bits & 0xF3);
          pad2_bits = (pad2_bits & 0x0C) | (pad_swap & 0xF3);
          pad_swap = 1;
        }
        if(GetAsyncKeyState('M') & 0x8000) reg4016 |= 0x04; // MIC, Insert Coin

        // Arkanoid Paddle
        if(ex_controller_type == EX_ARKANOID)
        {
          if(GetAsyncKeyState(VK_RBUTTON) & 0x8000) pad1_bits |= 0x04;
        }

        // Clazy Climber
        if(ex_controller_type == EX_CRAZY_CLIMBER)
        {
          pad1_bits = pad1_bits & 0x0F;
          pad2_bits = pad2_bits & 0x0F;
          if(GetAsyncKeyState('D') & 0x8000) pad1_bits |= 0x10;
          if(GetAsyncKeyState('A') & 0x8000) pad1_bits |= 0x20;
          if(GetAsyncKeyState('W') & 0x8000) pad1_bits |= 0x40;
          if(GetAsyncKeyState('S') & 0x8000) pad1_bits |= 0x80;
          if(GetAsyncKeyState('J') & 0x8000) pad2_bits |= 0x10;
          if(GetAsyncKeyState('G') & 0x8000) pad2_bits |= 0x20;
          if(GetAsyncKeyState('Y') & 0x8000) pad2_bits |= 0x40;
          if(GetAsyncKeyState('H') & 0x8000) pad2_bits |= 0x80;
        }

        // Hyper Shot
        if(ex_controller_type == EX_HYPER_SHOT)
        {
          if(pad1_bits & 0x01) reg4017 |= 0x02;
          if(pad1_bits & 0x02) reg4017 |= 0x04;
          if(pad2_bits & 0x01) reg4017 |= 0x08;
          if(pad2_bits & 0x02) reg4017 |= 0x10;
        }

        // Family Trainer, Pokkun Moguraa (start / select)
        if(ex_controller_type == EX_FAMILY_TRAINER_A ||
           ex_controller_type == EX_FAMILY_TRAINER_B ||
           ex_controller_type == EX_POKKUN_MOGURAA)
        {
          if(GetAsyncKeyState('T') & 0x8000) pad1_bits |= 0x08;
          if(GetAsyncKeyState('G') & 0x8000) pad1_bits |= 0x04;
        }

        // Power Pad (side A)
        if(ex_controller_type == EX_POWER_PAD_A)
        {
          if(GetAsyncKeyState('R') & 0x8000) reg4017s3 |= 0x02;
          if(GetAsyncKeyState('E') & 0x8000) reg4017s3 |= 0x01;
          if(GetAsyncKeyState('W') & 0x8000) reg4017s4 |= 0x02;
          if(GetAsyncKeyState('Q') & 0x8000) reg4017s4 |= 0x01;
          if(GetAsyncKeyState('F') & 0x8000) reg4017s3 |= 0x04;
          if(GetAsyncKeyState('D') & 0x8000) reg4017s3 |= 0x10;
          if(GetAsyncKeyState('S') & 0x8000) reg4017s3 |= 0x80;
          if(GetAsyncKeyState('A') & 0x8000) reg4017s4 |= 0x08;
          if(GetAsyncKeyState('V') & 0x8000) reg4017s3 |= 0x08;
          if(GetAsyncKeyState('C') & 0x8000) reg4017s3 |= 0x20;
          if(GetAsyncKeyState('X') & 0x8000) reg4017s3 |= 0x40;
          if(GetAsyncKeyState('Z') & 0x8000) reg4017s4 |= 0x04;
          if(GetAsyncKeyState('T') & 0x8000) pad1_bits |= 0x08;
          if(GetAsyncKeyState('G') & 0x8000) pad1_bits |= 0x04;
        }

        // Power Pad (side B)
        if(ex_controller_type == EX_POWER_PAD_B)
        {
          if(GetAsyncKeyState('Q') & 0x8000) reg4017s3 |= 0x02;
          if(GetAsyncKeyState('W') & 0x8000) reg4017s3 |= 0x01;
          if(GetAsyncKeyState('E') & 0x8000) reg4017s4 |= 0x02;
          if(GetAsyncKeyState('R') & 0x8000) reg4017s4 |= 0x01;
          if(GetAsyncKeyState('A') & 0x8000) reg4017s3 |= 0x04;
          if(GetAsyncKeyState('S') & 0x8000) reg4017s3 |= 0x10;
          if(GetAsyncKeyState('D') & 0x8000) reg4017s3 |= 0x80;
          if(GetAsyncKeyState('F') & 0x8000) reg4017s4 |= 0x08;
          if(GetAsyncKeyState('Z') & 0x8000) reg4017s3 |= 0x08;
          if(GetAsyncKeyState('X') & 0x8000) reg4017s3 |= 0x20;
          if(GetAsyncKeyState('C') & 0x8000) reg4017s3 |= 0x40;
          if(GetAsyncKeyState('V') & 0x8000) reg4017s4 |= 0x04;
          if(GetAsyncKeyState('T') & 0x8000) pad1_bits |= 0x08;
          if(GetAsyncKeyState('G') & 0x8000) pad1_bits |= 0x04;
        }
      }
    }

    // Arkanoid Paddle
    if(ex_controller_type == EX_ARKANOID)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      //if(data & 0x01)
      {
        if(GetAsyncKeyState(VK_LBUTTON) & 0x8000) reg4016 = 0x02;
        try
        {
          POINT pt;
          GetCursorPos(&pt);
          ScreenToClient(main_window_handle, &pt);
          if(NESTER_settings.nes.graphics.osd.double_size)
          {
            pt.x >>= 1;
          }
          if(pt.x > 176) pt.x = 176;
          if(pt.x < 32) pt.x = 32;
          uint8 px = 0xFF - (uint8)(0x52+172*(pt.x-32)/144);
          reg4017s1 |= (px & 0x01) << 7;
          reg4017s1 |= (px & 0x02) << 5;
          reg4017s1 |= (px & 0x04) << 3;
          reg4017s1 |= (px & 0x08) << 1;
          reg4017s1 |= (px & 0x10) >> 1;
          reg4017s1 |= (px & 0x20) >> 3;
          reg4017s1 |= (px & 0x40) >> 5;
          reg4017s1 |= (px & 0x80) >> 7;
        } catch(...) { }
      }
    }

    // Exciting Boxing
    if(ex_controller_type == EX_EXCITING_BOXING)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      if(data & 0x02)
      {
        if(!(GetAsyncKeyState('D') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('E') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('S') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('W') & 0x8000)) reg4017 |= 0x02;
      }
      else
      {
        if(!(GetAsyncKeyState('R') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('F') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('A') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('Q') & 0x8000)) reg4017 |= 0x02;
      }
    }

    // Family Basic keyboard (thanx rinao)
    if (ex_controller_type == EX_FAMILY_KEYBOARD)
    {
      if (data == 0x05)
      {
        kb_out = 0;
        kb_scan = 0;
      }
      else if (data  == 0x04)
      {
        kb_scan++;
        if (kb_scan > 9) kb_scan = 0;
        kb_out = 1 - kb_out;
      }
      else if (data == 0x06)
      {
        kb_out = 1 - kb_out;
      }
    }

    // Family Trainer (Side A) / Pokkun Moguraa
    if(ex_controller_type == EX_FAMILY_TRAINER_A ||
       ex_controller_type == EX_POKKUN_MOGURAA)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      if(!(data & 0x04))
      {
        if(!(GetAsyncKeyState('R') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('E') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('W') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('Q') & 0x8000)) reg4017 |= 0x02;
      }
      if(!(data & 0x02))
      {
        if(!(GetAsyncKeyState('F') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('D') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('S') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('A') & 0x8000)) reg4017 |= 0x02;
      }
      if(!(data & 0x01))
      {
        if(!(GetAsyncKeyState('V') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('C') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('X') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('Z') & 0x8000)) reg4017 |= 0x02;
      }
    }

    // Family Trainer (Side B)
    if(ex_controller_type == EX_FAMILY_TRAINER_B)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      if(!(data & 0x04))
      {
        if(!(GetAsyncKeyState('Q') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('W') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('E') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('R') & 0x8000)) reg4017 |= 0x02;
      }
      if(!(data & 0x02))
      {
        if(!(GetAsyncKeyState('A') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('S') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('D') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('F') & 0x8000)) reg4017 |= 0x02;
      }
      if(!(data & 0x01))
      {
        if(!(GetAsyncKeyState('Z') & 0x8000)) reg4017 |= 0x10;
        if(!(GetAsyncKeyState('X') & 0x8000)) reg4017 |= 0x08;
        if(!(GetAsyncKeyState('C') & 0x8000)) reg4017 |= 0x04;
        if(!(GetAsyncKeyState('V') & 0x8000)) reg4017 |= 0x02;
      }
    }

    // Mahjong Controller
    if(ex_controller_type == EX_MAHJONG)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      if((data & 0x06) == 0x02)
      {
        if(GetAsyncKeyState('J') & 0x8000) reg4017s1 |= 0x04;
        if(GetAsyncKeyState('H') & 0x8000) reg4017s1 |= 0x08;
        if(GetAsyncKeyState('G') & 0x8000) reg4017s1 |= 0x10;
        if(GetAsyncKeyState('F') & 0x8000) reg4017s1 |= 0x20;
        if(GetAsyncKeyState('D') & 0x8000) reg4017s1 |= 0x40;
        if(GetAsyncKeyState('S') & 0x8000) reg4017s1 |= 0x80;
      }
      else if((data & 0x06) == 0x04)
      {
        if(GetAsyncKeyState('A') & 0x8000) reg4017s1 |= 0x01;
        if(GetAsyncKeyState('U') & 0x8000) reg4017s1 |= 0x02;
        if(GetAsyncKeyState('Y') & 0x8000) reg4017s1 |= 0x04;
        if(GetAsyncKeyState('T') & 0x8000) reg4017s1 |= 0x08;
        if(GetAsyncKeyState('R') & 0x8000) reg4017s1 |= 0x10;
        if(GetAsyncKeyState('E') & 0x8000) reg4017s1 |= 0x20;
        if(GetAsyncKeyState('W') & 0x8000) reg4017s1 |= 0x40;
        if(GetAsyncKeyState('Q') & 0x8000) reg4017s1 |= 0x80;
      }
      else if((data & 0x06) == 0x06)
      {
        if(GetAsyncKeyState('M') & 0x8000) reg4017s1 |= 0x02;
        if(GetAsyncKeyState('N') & 0x8000) reg4017s1 |= 0x04;
        if(GetAsyncKeyState('B') & 0x8000) reg4017s1 |= 0x08;
        if(GetAsyncKeyState('V') & 0x8000) reg4017s1 |= 0x10;
        if(GetAsyncKeyState('C') & 0x8000) reg4017s1 |= 0x20;
        if(GetAsyncKeyState('X') & 0x8000) reg4017s1 |= 0x40;
        if(GetAsyncKeyState('Z') & 0x8000) reg4017s1 |= 0x80;
      }
    }

    // Oeka Kids Tablet
    if(ex_controller_type == EX_OEKAKIDS_TABLET)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      if(!(data & 0x01))
      {
        reg4017 = 0;
        tablet_data = 0;
        if(GetAsyncKeyState(VK_LBUTTON) & 0x8000) tablet_data |= 0x0001;
        try
        {
          POINT pt;
          GetCursorPos(&pt);
          ScreenToClient(main_window_handle, &pt);
          if(NESTER_settings.nes.graphics.osd.double_size)
          {
            pt.x >>= 1;
            pt.y >>= 1;
          }
          if(!NESTER_settings.nes.graphics.show_all_scanlines)
          {
            pt.y += 8;
          }
          if(pt.y > 48)
          {
            if(GetAsyncKeyState(VK_LBUTTON) & 0x8000) tablet_data |= 0x0001;
            tablet_data |= 0x0002;
          }
          else
          {
            if(GetAsyncKeyState(VK_LBUTTON) & 0x8000) tablet_data |= 0x0003;
          }
          pt.x = (int32)((240*pt.x)/256) +  8; if(pt.x < 0) pt.x = 0;
          pt.y = (int32)((256*pt.y)/240) - 12; if(pt.y < 0) pt.y = 0;
          tablet_data = tablet_data | (pt.x << 10) | (pt.y << 2);
        } catch(...) { }
      }
      else
      {
        if(!(tablet_pre_flag & 0x02) && (data & 0x02))
        {
          tablet_data <<= 1;
        }
        if(!(data & 0x02))
        {
          reg4017 = 0x04;
        }
        else
        {
          if(tablet_data & 0x40000)
          {
            reg4017 = 0x00;
          }
          else
          {
            reg4017 = 0x08;
          }
        }
      }
      tablet_pre_flag = data;
    }

    // Turbo File
    if(ex_controller_type == EX_TURBO_FILE)
    {
      reg4016 = reg4016s3 = reg4016s4 = 0;
      reg4017 = reg4017s1 = reg4017s3 = reg4017s4 = 0;
      if(!(data & 0x02))
      {
        tf_pointer = 0x00;
        tf_bit = 0x01;
      }
      if(data & 0x04)
      {
        tf_data[tf_bank*0x2000 + tf_pointer] &= 0xFF - tf_bit;
        tf_data[tf_bank*0x2000 + tf_pointer] |= tf_bit * (data & 0x01);
        tf_write = 1;
      }
      if((tf_pre_flag & 0x04) && !(data & 0x04))
      {
        if(tf_bit == 0x80)
        {
          tf_pointer = (tf_pointer + 1) & 0x1FFF;
          tf_bit = 0x01;
        }
        else
        {
          tf_bit <<= 1;
        }
      }
      if(tf_data[tf_bank*0x2000 + tf_pointer] & tf_bit)
      {
        reg4017 |= 0x04;
      }
      else
      {
        reg4017 &= 0xFB;
      }
      tf_pre_flag = data;
    }
  }
  else if(addr == 0x4017) // frame-IRQ
  {
    if(!frame_irq_disenabled)
    {
      frame_irq_enabled = data;
    }
  }
}

void NES::emulate_CPU_cycles(float num_cycles)
{
  uint32 cycle_deficit;

  ideal_cycle_count += num_cycles;
  cycle_deficit = ((uint32)ideal_cycle_count) - emulated_cycle_count;
  if(cycle_deficit > 0)
  {
    emulated_cycle_count += cpu->Execute(cycle_deficit);
    if(apu->SyncDMCRegister(cycle_deficit) && DPCM_IRQ)
    {
      cpu->DoIRQ();
    }
  }
}

// call every once in a while to avoid cycle count overflow
void NES::trim_cycle_counts()
{
  uint32 trim_amount;

  trim_amount = (uint32)floor(ideal_cycle_count);
  if(trim_amount > emulated_cycle_count) trim_amount = emulated_cycle_count;

  ideal_cycle_count  -= (float)trim_amount;
  emulated_cycle_count -= trim_amount;
}


void NES::Save_SaveRAM()
{
  // does the ROM use save ram?
  if(!ROM->has_save_RAM()) return;

  // has anything been written to Save RAM?
  for(uint32 i = 0; i < sizeof(SaveRAM); i++)
  {
    if(SaveRAM[i] != 0x00) break;
  }
  if(i < sizeof(SaveRAM))
  {
    FILE* fp = NULL;
    char fn[256];

    LOG("Saving Save RAM...");

    if( NESTER_settings.path.UseSramPath )
    {
      strcpy( fn, NESTER_settings.path.szSramPath );
      PathAddBackslash( fn );
    }
    else
      strcpy( fn, ROM->GetRomPath() );

    if( GetFileAttributes( fn ) == 0xFFFFFFFF )
      MKCreateDirectories( fn );

    strcat(fn, ROM->GetRomName());
    strcat(fn, ".sav");

    try
    {
      fp = fopen(fn, "wb");
      if(!fp) throw "can't open save RAM file";

      if(fwrite(SaveRAM, ROM->get_size_SaveRAM(), 1, fp) != 1)
        throw "can't open save RAM file";

      fclose(fp);
      LOG("Done." << endl);

    } catch(...) {
      LOG("can't save" << endl);
      if(fp) fclose(fp);
    }
  }
}

void NES::Load_SaveRAM()
{
  memset(SaveRAM, 0x00, sizeof(SaveRAM));

  // does the ROM use save ram?
  if(!ROM->has_save_RAM()) return;

  {
    FILE* fp = NULL;
    char fn[256];

    if( NESTER_settings.path.UseSramPath )
    {
      strcpy( fn, NESTER_settings.path.szSramPath );
      PathAddBackslash( fn );
    }
    else
      strcpy( fn, ROM->GetRomPath() );

    strcat(fn, ROM->GetRomName());
    strcat(fn, ".sav");

    try
    {
      fp = fopen(fn, "rb");
      if(!fp) throw "none found.";

      LOG("Loading Save RAM...");

      if(fread(SaveRAM, ROM->get_size_SaveRAM(), 1, fp) != 1)
      {
        LOG("error reading Save RAM file" << endl);
        throw "error reading Save RAM file";
      }

      fclose(fp);
      LOG("Done." << endl);

    } catch(...) {
      if(fp) fclose(fp);
    }
  }
}

void NES::Save_Disk()
{
  // disk system
  if(ROM->get_mapper_num() == 20)
  {
    FILE* fp = NULL;
    char fn[256];

    if( NESTER_settings.path.UseSramPath )
    {
      strcpy( fn, NESTER_settings.path.szSramPath );
      PathAddBackslash( fn );
    }
    else
      strcpy( fn, ROM->GetRomPath() );

    if( GetFileAttributes( fn ) == 0xFFFFFFFF )
      MKCreateDirectories( fn );

    strcat(fn, ROM->GetRomName());
    strcat(fn, ".sdk");

    boolean disk_changed = 0;
    for(uint32 i = 0; i < 0x40000; i++)
    {
      if(mapper->GetDiskChanged(i))
      {
        disk_changed = 1;
      }
    }

    if(disk_changed)
    {
      fp = fopen(fn, "wb");
      if(!fp) throw "can't open save DISK file";

      for(uint32 i = 0; i < 0x40000; i++)
      {
        if(mapper->GetDiskChanged(i))
        {
          fputc(mapper->GetDiskData(i), fp);
          fputc((i>> 0) & 0xFF, fp);
          fputc((i>> 8) & 0xFF, fp);
          fputc((i>>16) & 0xFF, fp);
        }
      }
      fclose(fp);
    }

    LOG("Done." << endl);
  }
}

void NES::Load_Disk()
{
  FILE* fp = NULL;
  char fn[256];

  if(ROM->get_mapper_num() == 20)
  {
    if( NESTER_settings.path.UseSramPath )
    {
      strcpy( fn, NESTER_settings.path.szSramPath );
      PathAddBackslash( fn );
    }
    else
      strcpy( fn, ROM->GetRomPath() );

    strcat(fn, ROM->GetRomName());
    strcat(fn, ".sdk");

    try
    {
      uint8 adr0,adr1,adr2,data;
      uint32 adr;

      for(adr = 0; adr < 0x40000; adr++)
      {
        mapper->SetDiskChanged(adr, 0);
      }

      fp = fopen(fn, "rb");
      if(!fp) throw "none found.";

      while((data = fgetc(fp)) != EOF)
      {
        adr0 = fgetc(fp);
        adr1 = fgetc(fp);
        adr2 = fgetc(fp);
        adr = adr0 | (adr1<<8) | (adr2<<16);
        mapper->SetDiskData(adr, data);
        mapper->SetDiskChanged(adr,1);
      }

      fclose(fp);
      LOG("Done." << endl);

    } catch(...) {
      if(fp) fclose(fp);
    }
  }
}

void NES::Save_TurboFile()
{
  if(ex_controller_type == EX_TURBO_FILE && tf_write)
  {
    FILE* fp = NULL;
    char fn[256];

    GetModuleFileName(NULL, fn, 256);
    PathRemoveFileSpec( fn );
    PathAddBackslash( fn );
    strcat(fn, "tb_file.dat");

    if((fp = fopen(fn, "wb")) != NULL)
    {
      fwrite(tf_data, 13*0x2000, 1, fp);
      fclose(fp);
    }
    tf_write = 0;
  }
}

void NES::Load_TurboFile()
{
  if(ex_controller_type == EX_TURBO_FILE)
  {
    FILE* fp = NULL;
    char fn[256];

    GetModuleFileName(NULL, fn, 256);
    PathRemoveFileSpec( fn );
    PathAddBackslash( fn );
    strcat(fn, "tb_file.dat");

    if((fp = fopen(fn, "rb")) != NULL)
    {
      fread(tf_data, 13*0x2000, 1, fp);
      fclose(fp);
    }
    tf_write = 0;
  }
}

boolean NES::loadState(const char* fn)
{
  return LoadSNSS(fn, this);
}

boolean NES::saveState(const char* fn)
{
  return SaveSNSS(fn, this);
}

void NES::calculate_palette()
{
  if(NESTER_settings.nes.graphics.calculate_palette == 1 && !use_vs_palette)
  {
    int x,z;
    float tint = ((float)NESTER_settings.nes.graphics.tint) / 256.0f;
    float hue = 332.0f + (((float)NESTER_settings.nes.graphics.hue - (float)0x80) * (20.0f / 256.0f));
    float s,y;
    int cols[16] = {0,240,210,180,150,120,90,60,30,0,330,300,270,0,0,0};
    float theta;
    float br1[4] = {0.5f, 0.75f, 1.0f, 1.0f};
    float br2[4] = {0.29f, 0.45f, 0.73f, 0.9f};
    float br3[4] = {0.0f, 0.24f, 0.47f, 0.77f};
    float r,g,b;

    for(x = 0; x <= 3; x++)
    {
      for(z = 0; z <= 15; z++)
      {
        s = tint;
        y = br2[x];
        if(z == 0)
        {
          s = 0;
          y = br1[x];
        }
        else if(z == 13)
        {
          s = 0;
          y = br3[x];
        }
        else if((z == 14) || (z == 15))
        {
          s = 0;
          y = 0;
        }

        theta = 3.14159265f * (((float)(cols[z] + hue)) / 180.0f);

        r = y + (s * (float)sin(theta));
        g = y - ((27.0f / 53.0f) * s * (float)sin(theta)) + ((10.0f / 53.0f) * s * (float)cos(theta));
        b = y - (s * (float)cos(theta));

        r = r * 256.0f;
        g = g * 256.0f;
        b = b * 256.0f;

        if(r > 255.0f) r = 255.0f;
        if(g > 255.0f) g = 255.0f;
        if(b > 255.0f) b = 255.0f;

        if(r < 0.0f) r = 0.0;
        if(g < 0.0f) g = 0.0;
        if(b < 0.0f) b = 0.0;

        NES_RGB_pal[(x*16) + z][0] = (uint8)r;
        NES_RGB_pal[(x*16) + z][1] = (uint8)g;
        NES_RGB_pal[(x*16) + z][2] = (uint8)b;
      }
    }
  }
  else if( NESTER_settings.nes.graphics.calculate_palette == 2 )
  {
    memset(NES_RGB_pal, 0x00, sizeof(NES_RGB_pal));
    if( FILE *pf = fopen( NESTER_settings.nes.graphics.szPaletteFile, "rb" ) )
    {
      fread( NES_RGB_pal, 1, sizeof(NES_RGB_pal), pf );
      fclose(pf);
    }
  }
  else
  {
    if( use_vs_palette )
    {
      memcpy(NES_RGB_pal, vs_palette, sizeof(NES_RGB_pal));
    }
    else
    {
      memcpy(NES_RGB_pal, NES_preset_palette, sizeof(NES_RGB_pal));
    }
  }

  if(NESTER_settings.nes.graphics.black_and_white)
  {
    int i;

    for(i = 0; i < NES_NUM_COLORS; i++)
    {
      uint8 Y;
      Y = (uint8)(((float)NES_RGB_pal[i][0] * 0.299) +
                  ((float)NES_RGB_pal[i][1] * 0.587) +
                  ((float)NES_RGB_pal[i][2] * 0.114));
      NES_RGB_pal[i][0] = Y;
      NES_RGB_pal[i][1] = Y;
      NES_RGB_pal[i][2] = Y;
    }
  }
}

uint8 NES::get_num_disks()
{
  return ROM->get_num_disks();
}
void NES::SetDiskSide(uint8 side)
{
  mapper->SetDiskSide( side );
}
uint8 NES::GetDiskSide()
{
  return mapper->GetDiskSide();
}

void NES::SetExControllerType(uint8 num)
{
  if(ex_controller_type != EX_TURBO_FILE && num != EX_TURBO_FILE)
  {
    ex_controller_type = num;
  }
}
uint8 NES::GetExControllerType()
{
  return ex_controller_type;
}
